#!/usr/bin/env python
# -*- coding: utf-8 -*

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import csv

import pandas as pd

from haystac.workflow.scripts.entrez_utils import entrez_esearch


def entrez_invalid_assemblies(assemblies, output):
    assemblies_file = pd.read_csv(assemblies, sep="\t")

    with open(output, "w") as fout:
        columns = ["species", "AccessionVersion"]
        w = csv.DictWriter(fout, columns, delimiter="\t")
        w.writeheader()

        # TODO why do this in a loop? much quicker to do this in batches of ENTREZ_MAX_UID
        #  for id_list in chunker(accessions, ENTREZ_MAX_UID)
        #       etree = entrez_efetch("assembly", id_list)
        for key, acc in assemblies_file.iterrows():
            # query the assembly database to confirm that this accession is still valid
            _, _, id_list = entrez_esearch("assembly", acc["AccessionVersion"] + ' AND "latest refseq"[filter]')

            if len(id_list) == 0:
                row = dict()
                row["species"] = acc["species"]
                row["AccessionVersion"] = acc["AccessionVersion"]
                w.writerow(row)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    entrez_invalid_assemblies(assemblies=snakemake.input[0], output=snakemake.output[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import csv
import os

import pandas as pd

from haystac.workflow.scripts.entrez_utils import entrez_efetch, ENTREZ_MAX_UID
from haystac.workflow.scripts.utilities import chunker


def entrez_taxonomy_query(nuccore_file, output_file):
    """
    Query the NCBI taxonomy database to get the taxa details for all the nuccore sequences.
    """

    assert os.stat(nuccore_file).st_size, f"The nuccore_query count file is empty {nuccore_file}"

    # load the unique list of taxa from the nuccore resultset
    accessions = [tax_id for tax_id in pd.read_csv(nuccore_file, sep="\t", usecols=["TaxId"], squeeze=True).unique()]

    with open(output_file, "w") as fout:
        columns = [
            "TaxId",
            "superkingdom",
            "phylum",
            "class",
            "order",
            "family",
            "genus",
            "species",
            "subspecies",
            "serovar",
        ]
        w = csv.DictWriter(fout, columns, delimiter="\t", extrasaction="ignore")
        w.writeheader()

        for id_list in chunker(accessions, ENTREZ_MAX_UID):

            # fetch the taxonomy metadata
            etree = entrez_efetch("taxonomy", id_list)

            for taxon in etree:
                row = dict()
                row["TaxId"] = taxon.find("TaxId").text

                # extract the taxonomic hierarchy
                for lineage in taxon.find("LineageEx"):
                    lineage_rank = lineage.find("Rank").text
                    if lineage_rank in columns:
                        row[lineage_rank] = lineage.find("ScientificName").text
                    elif lineage_rank == "no rank":
                        row["serovar"] = lineage.find("ScientificName").text

                # get the taxon rank and name
                taxon_rank = taxon.find("Rank").text
                taxon_name = taxon.find("ScientificName").text

                if taxon_rank in columns:
                    row[taxon_rank] = taxon_name

                if "serovar" not in row.keys():
                    if taxon_rank == "no rank":
                        row["serovar"] = taxon_name

                if row["species"] and not row.get("subspecies"):
                    row["subspecies"] = row["species"] + " ssp."

                if row["species"] and not row.get("serovar"):
                    row["serovar"] = ""

                w.writerow(row)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    entrez_taxonomy_query(
        nuccore_file=snakemake.input[0],
        output_file=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import csv

from haystac.workflow.scripts.entrez_utils import (
    entrez_esearch,
    entrez_esummary,
    entrez_xml_to_dict,
)
from haystac.workflow.scripts.utilities import print_error


def entrez_nuccore_query(query, output_file):
    """
    Query the NCBI nuccore database to get a list of sequence accessions and their metadata.
    """
    # execute the search
    key, webenv, id_list = entrez_esearch("nuccore", query)

    # check that there was at least one record found
    if len(id_list) == 0:
        print_error(f"The --query '{query}' returned no results")

    # fetch the results
    etree = entrez_esummary("nuccore", key, webenv)

    # convert the ElementTree into a a list of dicts
    data = entrez_xml_to_dict(etree)

    with open(output_file, "w") as fout:
        w = csv.DictWriter(fout, data[0].keys(), delimiter="\t")
        w.writeheader()
        w.writerows(data)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    entrez_nuccore_query(
        query=snakemake.config["query"],
        output_file=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import pandas as pd

from haystac.workflow.scripts.utilities import REGEX_BLACKLIST, print_error


def entrez_refseq_virus_create_files(
    config,
    input_file,
    viral_genomes_out,
):

    """Function to parse the refseq genomes report for viruses."""

    # read the file

    refseq_viruses = pd.read_csv(input_file, sep="\t")

    # drop duplicate species/strains

    refseq_viruses_rmdup = refseq_viruses[~refseq_viruses["#Organism/Name"].duplicated(keep="first")]

    viruses_unique = refseq_viruses_rmdup[["#Organism/Name", "Segmemts"]].copy()
    viruses_unique["Accession"] = ""

    # assign a segment accession code to a species.
    # 1 segment acc should be enough as all of them will be fetched through the assembly database.

    for index, row in viruses_unique.iterrows():

        seq_list = row["Segmemts"].split("; ")

        if len(seq_list) == 1 and ":" not in seq_list[0]:
            if "/" in seq_list[0]:
                row["Accession"] = seq_list[0].split("/")[0]
            elif "-" == seq_list[0]:
                continue
            else:
                row["Accession"] = seq_list[0]
        elif len(seq_list) == 1 and ":" in seq_list[0]:
            row["Accession"] = seq_list[0].split(":")[1].split("/")[0]
        elif len(seq_list) > 1:
            if "/" in seq_list[0].split(":")[1]:
                row["Accession"] = seq_list[0].split(":")[1].split("/")[0]
            else:
                row["Accession"] = seq_list[0].split(":")[1]
        else:
            print_error(seq_list)
            break

    # rename columns

    viruses = viruses_unique[["#Organism/Name", "Accession"]]
    viruses.rename(columns={"#Organism/Name": "species", "Accession": "AccessionVersion"}, inplace=True)

    # drop rows that have no accessions

    viruses = viruses[viruses["AccessionVersion"] != ""]

    # regex for species name

    viruses["species"] = viruses["species"].replace(REGEX_BLACKLIST, "_", regex=True)

    # check for duplicates from user input

    if config["sequences"] or config["accessions"]:
        user_inputs = []
        if os.path.isfile(config["sequences"]):
            custom_fasta_paths = pd.read_csv(
                config["sequences"],
                sep="\t",
                header=None,
                names=["species", "accession", "path"],
            )
            user_inputs.append(custom_fasta_paths)
        if os.path.isfile(config["accessions"]):
            custom_accessions = pd.read_csv(
                config["accessions"],
                sep="\t",
                header=None,
                names=["species", "accession"],
            )
            user_inputs.append(custom_accessions)

        for user_df in user_inputs:
            viruses = viruses[(~viruses["species"].isin(user_df["species"]))]

    # print the output to csv

    header = ["species", "AccessionVersion"]
    viruses.to_csv(viral_genomes_out, sep="\t", header=header, index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    entrez_refseq_virus_create_files(
        config=snakemake.config,
        input_file=snakemake.input[0],
        viral_genomes_out=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import time
from urllib.parse import urlencode
from xml.etree import ElementTree

import requests

from haystac.workflow.scripts.utilities import print_warning, print_error, get_smk_config

# base url of the Entrez web service
ENTREZ_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/"

# tool registration with NCBI, see https://www.ncbi.nlm.nih.gov/books/NBK25497/#chapter2.Frequency_Timing_and_Registrati
ENTREZ_TOOL = "haystac"
ENTREZ_EMAIL = "antonisdim41@gmail.com"

# maximum 1 request per X seconds
ENTREZ_WAIT_TIME = 1.1

# users that supply a valid API key can post 10 requests per second
ENTREZ_RATE_LOW = 3
ENTREZ_RATE_HIGH = 10

# maximum number of IDs that can be requested in one operation
ENTREZ_MAX_UID = 200

# maximum number of retries when we encounter NCBI server/network error
ENTREZ_MAX_ATTEMPTS = 2

# check if the error is because of NCBI [101=NewConnectionError, 104=ConnectionResetError, 110=TimeoutError,
# 111=ConnectionRefusedError, 500=InternalServerError]
ENTREZ_ERRORS = [101, 104, 110, 111, 500]


def entrez_request(action, params=None, attempt=1):
    """
    Helper function to ensure that we never exceed the rate limit.
    """
    params = params or dict()

    # tell NCBI which application is making these requests
    params["tool"] = ENTREZ_TOOL
    params["email"] = ENTREZ_EMAIL

    config = get_smk_config()

    if config.get("api_key"):
        # append the user specified api_key
        params["api_key"] = config["api_key"]

    url = ENTREZ_URL + action

    if len(params.get("id", [])) > ENTREZ_MAX_UID:
        print_error(f"List of Entrez IDs exceeds the maximum: {ENTREZ_MAX_UID}")

    if config.get("debug"):
        # turn into a get request
        print(params.items())
        get = dict(
            (key, value if isinstance(value, (str, int)) else ",".join(str(val) for val in value))
            for key, value in params.items()
        )
        print(f"{url}?{urlencode(get)}")

    # make the request
    r = requests.post(url, params)

    # enforce the rate limit (even when the request failed)
    time.sleep(ENTREZ_WAIT_TIME)

    if not r.ok:
        if r.status_code in ENTREZ_ERRORS and attempt < ENTREZ_MAX_ATTEMPTS:
            return entrez_request(action, params, attempt + 1)
        else:
            r.raise_for_status()

    return r


def entrez_esearch(database, query):
    """
    Execute an Entrez esearch query and return the search keys
    """
    r = entrez_request("esearch.fcgi", {"db": database, "term": query, "usehistory": "y"})

    # parse the XML result
    etree = ElementTree.XML(r.text)

    # get the search keys
    key = etree.find("QueryKey").text
    webenv = etree.find("WebEnv").text
    id_list = [id_node.text for id_node in etree.findall(".//Id")]

    return key, webenv, id_list


def entrez_esummary(database, query_key, webenv):
    """
    Fetch the Entrez esummary records for an esearch query.
    """
    r = entrez_request("esummary.fcgi", {"db": database, "query_key": query_key, "WebEnv": webenv})

    return ElementTree.XML(r.text)


def entrez_efetch(database, id_list):
    """
    Fetch the Entrez records for a list of IDs.
    """
    r = entrez_request("efetch.fcgi", {"db": database, "id": id_list})

    return ElementTree.XML(r.text)


def entrez_assembly_ftp(accession, force=False):
    """
    Get an NCBI ftp url from the assembly database.
    """

    # find out if we are looking for a virus to apply the correct filter
    filter_condition = ' AND "latest"[filter] NOT suppressed*'

    config = get_smk_config()

    if config.get("refseq_rep") and config["refseq_rep"] == "viruses":
        # append the virus filter
        filter_condition += " AND viruses[filter] "

    # query the assembly database to get the latest assembly for this accession code
    key, webenv, id_list = entrez_esearch(
        "assembly",
        accession + filter_condition,
    )

    if len(id_list) > 1:
        # should never happen, but...
        msg = f"Multiple assembly accessions found for '{accession}': {id_list}. "

        # if force-accessions is true pick the largest int value, assuming it is also the altest
        if force:
            msg += f"Using assembly: id pair '{accession}': '{max([int(id_num) for id_num in id_list])}'"
            id_list = [str(max([int(id_num) for id_num in id_list]))]
            print_warning(msg)

        # if not raise an error
        else:
            msg += (
                f"Either consider using the `--force-accessions` flag for the largest ID to be picked, "
                f"or the `--exclude-accessions` flag to remove accession '{accession}' from this query."
            )
            print_error(msg)

    elif len(id_list) == 0:
        # no entry in the assembly database for this accession code
        return ""

    # fetch the summary record for the assembly
    r = entrez_request("esummary.fcgi", {"db": "assembly", "id": id_list})

    # parse the XML result
    etree = ElementTree.XML(r.text)

    # check if the assembly is anomalous
    anomalous = [reason.text for reason in etree.findall(".//Anomalous/Property")]

    if anomalous:
        message = (
            f"Assembly '{accession}' has been marked as anomalous for the following reasons: "
            f"'{'; '.join(anomalous)}'"
        )

        if force:
            print_warning(message)
        else:
            print_error(message)

    refseq = etree.find(".//FtpPath_RefSeq")
    genbank = etree.find(".//FtpPath_GenBank")

    # preference RefSeq URLs over GenBank URLs
    if refseq is not None and refseq.text not in ["", None]:
        ftp_stub = refseq.text
    elif genbank is not None and genbank.text not in ["", None]:
        ftp_stub = genbank.text
    else:
        return ""

    # append the fasta filename
    ftp_url = os.path.join(ftp_stub, os.path.basename(ftp_stub) + "_genomic.fna.gz")

    return ftp_url


def entrez_range_accessions(accession, first, last):
    """
    Return a range between two accession codes (e.g. ABC001 -> ABC005)
    """
    # sanity check that the items are equal length and in order
    assert len(first) == len(last) and first < last

    # find the index of the first difference
    idx = [i for i in range(len(first)) if first[i] != last[i]][0]
    pad = len(first) - idx

    try:
        # return the range
        return [f"{first[:idx]}{str(item).zfill(pad)}" for item in range(int(first[idx:]), int(last[idx:]) + 1)]
    except ValueError:
        print_error(f"Could not resolve the accession range '{first}-{last}' for master record '{accession}'")


def entrez_xml_to_dict(etree):
    """
    Convert an ElementTree object into a list of dicts.
    """
    data = []

    for row_node in etree:
        row = {}
        for col_node in row_node:
            col = col_node.attrib.get("Name", col_node.tag)
            row[col] = col_node.text

        data.append(row)

    return data


def entrez_find_replacement_accession(accession):
    """
    If the updated version of an assembly accession if possible
    """

    # try getting a new accession for the master record
    accession_new = accession[:-1] + "0000000"

    try:
        # send a request and see if we get back an xml result
        r = entrez_request(
            "efetch.fcgi",
            {"db": "nuccore", "id": accession_new, "rettype": "gb", "retmode": "xml"},
        )

    except requests.exceptions.HTTPError:
        print_error(f"Could not find either the GenBank record for '{accession}' or an alternative accession")

    # noinspection PyUnboundLocalVariable
    etree = ElementTree.XML(r.text)
    replacement = etree.find(".//GBSeq_accession-version")

    # check that the new accession is a WGS project
    keywords = [keyword.text.lower() for keyword in etree.findall(".//GBKeyword")]

    if replacement is not None and "wgs" in keywords:
        print_warning(f"Replacing the superseded WGS accession '{accession}' with '{replacement.text}'")
        return replacement.text
    else:
        print_error(
            f"Could not find either the GenBank record for '{accession}' or a valid alternative accession. "
            f"Please consider using the `--exclude-accessions` flag to remove accession '{accession}' from this query."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os

import pandas as pd


def coverage_chi2_contingency_test(cov_file, taxon_fasta_idx, taxon, outfile):
    """
    Function that calculates the pvalue of the coverage. Testing if there was clustering bias during the
    sequencing/the reads that contribute to the identification/abundance of a species come only
    from a specific genomic region
    """

    assert os.stat(cov_file).st_size, f"The file with the coverage stats {cov_file} is empty."

    cov_stats = pd.read_csv(cov_file, sep="\t", names=["all_positions", "observed", "expected"])

    all_positions = cov_stats["all_positions"].iloc[0]
    ref_bases_cov = cov_stats["observed"].iloc[0]
    total_bases_cov = cov_stats["expected"].iloc[0]

    # calculate coverage
    coverage = total_bases_cov / all_positions

    # calculate the fraction of the genome that is covered
    taxon_seqlen = genome_sizes(taxon_fasta_idx)
    fraction_ref_cov = ref_bases_cov / taxon_seqlen

    # calculate an evenness of coverage metric valid only for data < 1x. If less than 10 / 1 that is good.
    cov_evenness = coverage / fraction_ref_cov

    with open(outfile, "w") as outhandle:
        print(taxon, ref_bases_cov, total_bases_cov, coverage, fraction_ref_cov, cov_evenness, file=outhandle, sep="\t")


def genome_sizes(taxon_fasta_idx):
    faidx = pd.read_csv(
        taxon_fasta_idx,
        sep="\t",
        names=["Name", "Length", "Offset", "Linebases", "Linewidth"],
    )

    taxon_seq_len = faidx["Length"].sum()

    return taxon_seq_len


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    coverage_chi2_contingency_test(
        cov_file=snakemake.input[0],
        taxon_fasta_idx=snakemake.input[1],
        taxon=snakemake.wildcards.orgname,
        outfile=snakemake.output[0],
    )



#! /usr/bin/env python
"""
Execution script for snakemake workflows.
"""
__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import argparse
import datetime
import os
import shutil
import sys
from multiprocessing import cpu_count
from snakemake.dag import Batch

import snakemake
import yaml
from psutil import virtual_memory

from haystac import __version__
from haystac.workflow.scripts.entrez_utils import (
    ENTREZ_RATE_LOW,
    ENTREZ_RATE_HIGH,
)
from haystac.workflow.scripts.utilities import (
    ValidationError,
    ArgumentCustomFormatter,
    WritablePathType,
    FloatRangeType,
    IntRangeType,
    BoolType,
    JsonType,
    SequenceFileType,
    AccessionFileType,
    SraAccessionType,
    NuccoreQueryType,
    CheckExistingConfig,
    BatchType,
    FastqFile,
    PE,
    COLLAPSED,
    SE,
    md5,
    print_warning,
    print_error,
)

os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

MAX_CPU = cpu_count()
MAX_MEM_MB = int(virtual_memory().total / 1024 ** 2)

CODE_DIR = os.path.abspath(os.path.dirname(__file__))
SNAKE_DIR = ".snakemake"

CONFIG_DEFAULT = f"{CODE_DIR}/config/config.yaml"
CONFIG_USER = os.path.abspath(os.path.expanduser("~/.haystac/config.yaml"))
CONFIG_RUNTIME = f"{SNAKE_DIR}/config.yaml"

COMMANDS = ["config", "database", "sample", "analyse"]

DATABASE_MODES = ["fetch", "index", "build"]
ANALYSIS_MODES = ["filter", "align", "likelihoods", "probabilities", "abundances", "reads"]
TAXONOMIC_RANKS = ["genus", "species", "subspecies", "serotype"]
REFSEQ_TABLES = ["prokaryote_rep", "viruses", "eukaryotes"]

# number to times to retry a rule that failed the first time
RESTART_TIMES = 2


class Haystac(object):
    """
    Command-line interface for running `haystac`
    """

    def __init__(self):
        parser = argparse.ArgumentParser(
            usage="""haystac <command> [<args>]

The haystac commands are:
   config         Configuration options
   database       Build a database of target species
   sample         Prepare a sample for analysis
   analyse        Analyse a sample against a database
   
""",
        )
        parser.add_argument("command", choices=COMMANDS, help="Command to run")

        # get the CLI arguments
        args = self._parse_args(parser, level=1)

        # load the config files
        self._load_config()

        try:
            # call the class method with the name given by `command`
            reval = getattr(self, args.command)()
            exit(reval)

        except ValidationError as error:
            print_error(f"{error}")

    @staticmethod
    def _parse_args(parser, level=1):
        """
        Handle the nested-subcommand CLI interface.
        """

        # if we don't have enough arguments, then print the help
        if len(sys.argv) <= level:
            parser.print_help()
            parser.exit()

        if {"-v", "--version"}.intersection(sys.argv):
            print(__version__)
            parser.exit()

        # slice the CLI arguments
        argv = sys.argv[1:2] if level == 1 else sys.argv[level:]

        return parser.parse_args(argv)

    def _load_config(self):
        """
        Load the default and user space config files.
        """

        # load the default config
        with open(CONFIG_DEFAULT) as fin:
            self.config_default = yaml.safe_load(fin)

            # resolve relative paths
            self.config_default["cache"] = os.path.abspath(os.path.expanduser(self.config_default["cache"]))

        try:
            # load the user config
            with open(CONFIG_USER) as fin:
                self.config_user = yaml.safe_load(fin)

        except FileNotFoundError:
            # make the config folder (if necessary)
            os.makedirs(os.path.dirname(CONFIG_USER), exist_ok=True)

            self.config_user = dict()

        # maximum concurrent Entrez requests
        self.max_entrez_requests = ENTREZ_RATE_HIGH if self.config_user.get("api_key") else ENTREZ_RATE_LOW

        # merge the config dictionaries (giving precedence to the config_user)
        self.config_merged = {**self.config_default, **self.config_user}

    def config(self):
        """
        Configuration options
        """
        # noinspection PyTypeChecker
        parser = argparse.ArgumentParser(
            prog="haystac config",
            description="Configuration options",
            formatter_class=ArgumentCustomFormatter,
            add_help=False,
        )

        optional = parser.add_argument_group("Optional arguments")

        # add the help option manually so we can control where it is shown in the menu
        optional.add_argument("-h", "--help", action="help", help="Show this help message and exit")
        optional.add_argument("-v", "--version", action="help", help="Print the version number and exit")

        optional.add_argument(
            "--cache",
            help=f"Cache folder for storing genomes downloaded from NCBI and other shared data "
            f"(default: {self.config_default['cache']})",
            metavar="<path>",
            type=WritablePathType(),
            default=argparse.SUPPRESS,
        )

        optional.add_argument(
            "--clear-cache",
            help="Clear the contents of the cache folder, and delete the folder itself",
            action="store_true",
        )

        optional.add_argument(
            "--api-key",
            help=f"Personal NCBI API key (increases max concurrent requests from {ENTREZ_RATE_LOW} to "
            f"{ENTREZ_RATE_HIGH}, https://www.ncbi.nlm.nih.gov/account/register/)",
            metavar="<code>",
            default=argparse.SUPPRESS,
        )

        optional.add_argument(
            "--use-conda",
            help=f"Use conda as a package manger (default: {self.config_default['use_conda']})",
            type=BoolType(),
            metavar="<bool>",
            default=argparse.SUPPRESS,
        )

        # get the CLI arguments
        args = self._parse_args(parser, level=2)

        # get the user choices
        for key, value in vars(args).items():
            self.config_user[key] = value

        # resolve relative paths
        if self.config_user.get("cache"):
            self.config_user["cache"] = os.path.abspath(os.path.expanduser(self.config_user["cache"]))

        # clear the cache directory if specified
        if self.config_user.get("clear_cache"):
            shutil.rmtree(self.config_user["cache"])
            # delete the clear_cache key, so that the cache is not cleared every time unless the user specifies it
            del self.config_user["clear_cache"]

        # save the user config
        with open(CONFIG_USER, "w") as fout:
            yaml.safe_dump(self.config_user, fout, default_flow_style=False)

    def database(self):
        """
        Build a database of target species
        """
        # noinspection PyTypeChecker
        parser = argparse.ArgumentParser(
            prog="haystac database",
            description="Build a database of target species",
            formatter_class=ArgumentCustomFormatter,
            add_help=False,
        )

        required = parser.add_argument_group("Required arguments")

        required.add_argument(
            "--mode",
            help="Database creation mode for haystac [%(choices)s]",
            metavar="<mode>",
            choices=DATABASE_MODES,
            required=True,
        )

        required.add_argument(
            "--output",
            help="Path to the database output directory",
            metavar="<path>",
            dest="db_output",
            type=WritablePathType(),
            required=True,
        )

        choice = parser.add_argument_group("Required choice")

        choice.add_argument(
            "--query",
            help="Database query in the NCBI query language. "
            "Please refer to the documentation for assistance with constructing a valid query.",
            metavar="<query>",
            type=NuccoreQueryType(),
        )

        choice.add_argument(
            "--query-file",
            help="File containing a database query in the NCBI query language.",
            metavar="<path>",
            type=NuccoreQueryType(),
        )

        choice.add_argument(
            "--accessions-file",
            dest="accessions",
            help="Tab delimited file containing one record per row: the name of the taxon, "
            "and a valid NCBI accession code from the nucleotide, assembly or WGS databases.",
            metavar="<path>",
            type=AccessionFileType(),
        )

        choice.add_argument(
            "--sequences-file",
            dest="sequences",
            help="Tab delimited file containing one record per row: the name of the taxon, a user defined "
            "accession code, and the path to the fasta file (optionally compressed).",
            metavar="<path>",
            type=SequenceFileType(),
        )

        choice.add_argument(
            "--refseq-rep",
            help="Use one of the RefSeq curated tables to construct a DB. Includes all prokaryotic species "
            "(excluding strains) from the representative RefSeq DB, or all the species and strains from the "
            "viruses DB, or all the species and subspecies from the eukaryotes DB. "
            "If multiple accessions exist for a given species/strain, the first pair of species/accession is kept. "
            "Available RefSeq tables to use [%(choices)s].",
            choices=REFSEQ_TABLES,
            metavar="<table>",
            default=None,
        )

        optional = parser.add_argument_group("Optional arguments")

        optional.add_argument(
            "--force-accessions",
            help="Disable validation checks for 'anomalous' assembly flags in NCBI",
            action="store_true",
        )

        optional.add_argument(
            "--exclude-accessions",
            help="List of NCBI accessions to exclude.",
            metavar="<accession>",
            nargs="+",
            default=[],
        )

        optional.add_argument(
            "--resolve-accessions",
            help="Pick the first accession when two accessions for a taxon can be found in user provided input files",
            action="store_true",
        )

        optional.add_argument(
            "--bowtie2-scaling",
            help="Rescaling factor to keep the bowtie2 mutlifasta index below the maximum memory limit",
            type=FloatRangeType(0, 100),
            metavar="<float>",
            default=self.config_default["bowtie2_scaling"],
        )

        optional.add_argument(
            "--bowtie2-threads",
            help="Number of threads bowtie2 will use to index every individual genome in the database",
            type=IntRangeType(1, MAX_CPU),
            metavar="<int>",
            dest="bowtie2_threads_db",
            default=self.config_default["bowtie2_threads_db"],
        )

        optional.add_argument(
            "--rank",
            help="Taxonomic rank to perform the identifications on [%(choices)s]",
            metavar="<rank>",
            choices=TAXONOMIC_RANKS,
            default=self.config_default["rank"],
        )

        optional.add_argument(
            "--genera",
            help="List of genera to restrict the abundance calculations.",
            metavar="<genus>",
            nargs="+",
            default=None,
        )

        optional.add_argument(
            "--mtDNA",
            help="For eukaryotes, download mitochondrial genomes only. "
            "Not to be used with --refseq-rep or queries containing prokaryotes",
            action="store_true",
        )

        optional.add_argument(
            "--seed",
            help="Random seed for database indexing",
            metavar="<int>",
            type=int,
            default=self.config_default["seed"],
        )

        optional.add_argument(
            "--batch",
            help="Batch number for large`haystac database` workflows (e.g. --batch index_all_accessions=1/3). "
            "You will need to execute all batches before haystac is able to finish its workflow to the end.",
            metavar="<str>",
            type=BatchType(),
            default=None,
        )

        # add the common arguments
        self._common_arguments(parser)

        # get the CLI arguments
        args = self._parse_args(parser, level=2)

        # must specify at least one source for the database
        if args.mode != "index" and not (
            args.refseq_rep or args.query or args.query_file or args.accessions or args.sequences
        ):
            raise ValidationError(
                "Please specify at least one of --refseq-rep, --query, --query-file, --accessions or --sequences"
            )

        if args.mtDNA and args.refseq_rep:
            raise ValidationError("Please specify either `--mtDNA` or `--refseq-rep` but not both.")

        if args.query and args.query_file:
            raise ValidationError("Please specify either `--query <query>` or `--query-file <path>` but not both.")

        if args.query_file:
            # load the query file
            with open(args.query_file) as fin:
                args.query = fin.read().strip()

        # warn the user in case he is selecting only one thread for bt2 accession indexing
        if args.bowtie2_threads_db == 1:
            print_warning(
                "Please note that when bowtie2 uses only 1 thread for indexing its memory usage tends "
                "to be higher than when it is multithreaded. That means that HAYSTAC could end up using "
                "more memory than it is desired. Please be cautious when using only 1 thread "
                "for indexing individual genomes."
            )

        db_original = args.db_output

        # resolve relative paths
        args.db_output = os.path.abspath(os.path.expanduser(args.db_output))

        # cast all `None` paths as "" or else smk complains when parsing the rules
        args.query_file = args.query_file or ""
        args.accessions = args.accessions or ""
        args.sequences = args.sequences or ""

        # add all command line options to the merged config
        # noinspection PyDictCreation
        config = {**self.config_merged, **vars(args)}

        # store the md5 checksums for the database user input files
        config["query_file_md5"] = md5(args.query_file) if args.query_file else ""
        config["accessions_md5"] = md5(args.accessions) if args.query_file else ""
        config["sequences_md5"] = md5(args.sequences) if args.query_file else ""

        config_fetch = os.path.join(args.db_output, "database_fetch_config.yaml")
        config_build = os.path.join(args.db_output, "database_build_config.yaml")

        if args.refseq_rep:
            if not args.force_accessions:
                print_warning("Automatically setting `--force-accessions` in `--refseq-rep` mode")

            config["force_accessions"] = True

        target_list = list()

        if args.mode == "fetch":
            if args.query or args.accessions or args.sequences or args.refseq_rep or args.query_file:
                target_list.append("db_taxa_accessions.tsv")

            CheckExistingConfig(config_fetch, config)

            with open(config_fetch, "w") as fout:
                yaml.safe_dump(config, fout, default_flow_style=False)

        elif args.mode == "index":
            target_list.append("bowtie/bowtie_index.done")
            target_list.append("idx_database.done")

            try:
                with open(config_fetch, "r") as fin:
                    # config = {**yaml.safe_load(fin), **config}
                    fetch_args = {**yaml.safe_load(fin)}
                    fetch_args.update((k, v) for k, v in config.items() if v is not None)
                    config = fetch_args

                    CheckExistingConfig(config_fetch, config)
            except FileNotFoundError:
                raise ValidationError(
                    "Please run haystac `database --mode fetch` before attempting to index the database."
                )

        elif args.mode == "build":
            target_list.append("idx_database.done")
            target_list.append("bowtie/bowtie_index.done")
            target_list.append("db_taxa_accessions.tsv")

            if os.path.exists(config_fetch):
                raise ValidationError(
                    "Please run haystac `database --mode index` as the database has already been fetched."
                )

            if os.path.exists(config_build):
                CheckExistingConfig(config_build, config)

            with open(config_build, "w") as fout:
                yaml.safe_dump(config, fout, default_flow_style=False)

        target_list = [os.path.join(args.db_output, target) for target in target_list]

        # run the `haystac` workflow
        exit_code = self._run_snakemake("database", args, config, target_list)

        if args.mode == "fetch" and exit_code == 0 and not args.unlock:
            print(f"Please run `haystac database --mode index --output {db_original}` after this step.")

        return exit_code

    def sample(self):
        """
        Prepare a sample for analysis
        """
        # noinspection PyTypeChecker
        parser = argparse.ArgumentParser(
            prog="haystac sample",
            description="Prepare a sample for analysis",
            formatter_class=ArgumentCustomFormatter,
            add_help=False,
        )

        required = parser.add_argument_group("Required arguments")

        required.add_argument(
            "--output",
            help="Path to the sample output directory",
            metavar="<path>",
            dest="sample_output_dir",
            type=WritablePathType(),
            required=True,
        )

        choice = parser.add_argument_group("Required choice")

        choice.add_argument(
            "--fastq",
            help="Single-end fastq input file (optionally compressed).",
            metavar="<path>",
            type=FastqFile(),
        )

        choice.add_argument(
            "--fastq-r1",
            help="Paired-end forward strand (R1) fastq input file.",
            metavar="<path>",
            type=FastqFile(),
        )

        choice.add_argument(
            "--fastq-r2",
            help="Paired-end reverse strand (R2) fastq input file.",
            metavar="<path>",
            type=FastqFile(),
        )

        choice.add_argument(
            "--sra",
            help="Download fastq input from the SRA database",
            type=SraAccessionType(),
            metavar="<accession>",
        )

        optional = parser.add_argument_group("Optional arguments")

        optional.add_argument(
            "--collapse",
            help="Collapse overlapping paired-end reads, e.g. for aDNA",
            type=BoolType(),
            metavar="<bool>",
            default=self.config_default["collapse"],
        )

        optional.add_argument(
            "--trim-adapters",
            help="Automatically trim sequencing adapters from fastq input",
            type=BoolType(),
            metavar="<bool>",
            default=self.config_default["trim_adapters"],
        )

        # add the common arguments
        self._common_arguments(parser)

        # get the CLI arguments
        args = self._parse_args(parser, level=2)

        # must specify exactly one source for the sample
        if bool(args.fastq) + (bool(args.fastq_r1) and bool(args.fastq_r2)) + bool(args.sra) != 1:
            raise ValidationError("Please specify either --sra or --fastq, or both --fastq-r1 and --fastq-r2.")

        if (bool(args.fastq) and bool(args.fastq_r1)) or (bool(args.fastq) and bool(args.fastq_r2)):
            raise ValidationError(
                "Please specify either --fastq, or both --fastq-r1 and --fastq-r2. "
                "Any combination of these flags is not valid"
            )

        if bool(args.sra) and (bool(args.fastq_r1) or bool(args.fastq) or bool(args.fastq_r2)):
            raise ValidationError(
                "Please specify either --sra or --fastq, or both --fastq-r1 and --fastq-r2. "
                "Any combination of these flags is not valid"
            )

        if args.collapse and (not (args.fastq_r1 and args.fastq_r2) and (bool(args.sra) and args.sra[1] != "paired")):
            raise ValidationError("Collapse can only be used with --fastq-r1 and --fastq-r2 or sra PE data.")

        if args.collapse and not args.trim_adapters:
            raise ValidationError("Collapse can only be used with `--trim-adapters True`.")

        # resolve relative paths
        args.sample_output_dir = os.path.abspath(os.path.expanduser(args.sample_output_dir))

        # cast all `None` paths as "" or ele smk complains when parsing the rules
        args.fastq = args.fastq or ""
        args.fastq_r1 = args.fastq_r1 or ""
        args.fastq_r2 = args.fastq_r2 or ""

        # add all command line options to the merged config
        config = {**self.config_merged, **vars(args)}

        if not (args.fastq_r1 and args.fastq_r2):
            config["read_mode"] = SE
        if (args.fastq_r1 and args.fastq_r2) and args.collapse:
            config["read_mode"] = COLLAPSED
        if (args.fastq_r1 and args.fastq_r2) and not args.collapse:
            config["read_mode"] = PE

        # use the --output folder to name the sample
        config["sample_prefix"] = os.path.basename(config["sample_output_dir"].rstrip("/"))

        if args.sra:
            config["sra"], config["layout"] = args.sra

            # query the SRA to see if this is a paired-end library or not
            if config["layout"] == "paired":
                if config["collapse"]:
                    config["read_mode"] = COLLAPSED
                else:
                    config["read_mode"] = PE

                config["fastq_r1"] = config["sample_output_dir"] + f"/sra_data/PE/{config['sra']}_R1.fastq.gz"
                config["fastq_r2"] = config["sample_output_dir"] + f"/sra_data/PE/{config['sra']}_R1.fastq.gz"

            else:
                config["read_mode"] = SE
                config["fastq"] = config["sample_output_dir"] + f"/sra_data/SE/{config['sra']}.fastq.gz"

        target_list = list()
        target_list.append(f"fastq_inputs/meta/{config['sample_prefix']}.size")

        if config["trim_adapters"]:
            if config["read_mode"] == PE:
                target_list.append(f"fastq_inputs/{config['read_mode']}/{config['sample_prefix']}_R1_adRm.fastq.gz")
            elif config["read_mode"] == COLLAPSED:
                target_list.append(f"fastq_inputs/{config['read_mode']}/{config['sample_prefix']}_adRm.fastq.gz")
            elif config["read_mode"] == SE:
                target_list.append(f"fastq_inputs/{config['read_mode']}/{config['sample_prefix']}_adRm.fastq.gz")

        config_sample = os.path.join(args.sample_output_dir, "sample_config.yaml")

        if os.path.exists(config_sample):
            CheckExistingConfig(config_sample, config)

        with open(config_sample, "w") as fout:
            yaml.safe_dump(config, fout, default_flow_style=False)

        target_list = [os.path.join(args.sample_output_dir, target) for target in target_list]

        return self._run_snakemake("sample", args, config, target_list)

    def analyse(self):
        """
        Analyse a sample against a database
        """
        # noinspection PyTypeChecker
        parser = argparse.ArgumentParser(
            prog="haystac analyse",
            description="Analyse a sample against a database",
            formatter_class=ArgumentCustomFormatter,
            add_help=False,
        )

        required = parser.add_argument_group("Required arguments")

        required.add_argument(
            "--mode",
            help="Analysis mode for the selected sample [%(choices)s]",
            metavar="<mode>",
            choices=ANALYSIS_MODES,
            required=True,
        )

        required.add_argument(
            "--database",
            help="Path to the database output directory",
            metavar="<path>",
            dest="db_output",
            type=WritablePathType(),
            required=True,
        )

        required.add_argument(
            "--sample",
            help="Path to the sample output directory",
            metavar="<path>",
            dest="sample_output_dir",
            type=WritablePathType(),
            required=True,
        )

        required.add_argument(
            "--output",
            help="Path to the analysis output directory",
            metavar="<path>",
            dest="analysis_output_dir",
            type=WritablePathType(),
            required=True,
        )

        optional = parser.add_argument_group("Optional arguments")

        optional.add_argument(
            "--genera",
            help="List of genera to restrict the abundance calculations",
            metavar="<genus>",
            nargs="+",
            default=[],
        )

        optional.add_argument(
            "--min-prob",
            help="Minimum posterior probability to assign an aligned read to a given species (default: 0.50)",
            type=FloatRangeType(0, 100),
            metavar="<float>",
            default=argparse.SUPPRESS,
        )

        optional.add_argument(
            "--mismatch-probability",
            help="Base mismatch probability (default: 0.15)",
            type=FloatRangeType(0.01, 0.5),
            metavar="<float>",
            default=argparse.SUPPRESS,
        )

        optional.add_argument(
            "--bowtie2-threads",
            help="Number of threads bowtie2 will use to align a sample against every individual genome in the database",
            type=IntRangeType(1, MAX_CPU),
            metavar="<int>",
            dest="bowtie2_threads_aln",
            default=self.config_default["bowtie2_threads_aln"],
        )

        optional.add_argument(
            "--batch",
            help="Batch number for large`haystac analyse` workflows (e.g. --batch align_all_accessions=1/3). "
            "You will need to execute all batches before haystac is able to finish its workflow to the end.",
            metavar="<str>",
            type=BatchType(),
            default=None,
        )

        optional.add_argument(
            "--mapdamage",
            help="Perform mapdamage analysis for ancient samples (default: False)",
            type=BoolType(),
            metavar="<bool>",
            default=argparse.SUPPRESS,
        )

        optional.add_argument(
            "--aDNA",
            help="Set new flag defaults for aDNA sample analysis: "
            "mismatch-probability=0.20, min-prob=0.5, mapdamage=True",
            action="store_true",
        )

        # add the common arguments
        self._common_arguments(parser)

        # get the CLI arguments
        args = self._parse_args(parser, level=2)

        # resolve relative paths
        args.db_output = os.path.abspath(os.path.expanduser(args.db_output))
        args.sample_output_dir = os.path.abspath(os.path.expanduser(args.sample_output_dir))
        args.analysis_output_dir = os.path.abspath(os.path.expanduser(args.analysis_output_dir))

        config_fetch = os.path.join(args.db_output, "database_fetch_config.yaml")
        config_build = os.path.join(args.db_output, "database_build_config.yaml")
        config_sample = os.path.join(args.sample_output_dir, "sample_config.yaml")

        try:
            # load the database config
            database_config_file = config_fetch if os.path.exists(config_fetch) else config_build
            with open(database_config_file, "r") as fin:
                database_config = yaml.safe_load(fin)
        except FileNotFoundError:
            raise ValidationError(
                "The database has not been built correctly. Please rebuild the database using `haystac database`"
            )

        try:
            # load the sample config
            with open(config_sample, "r") as fin:
                sample_config = yaml.safe_load(fin)
        except FileNotFoundError:
            raise ValidationError(
                "The sample has not been prepared correctly. Please pre-process the samples using `haystac sample`"
            )

        # Check that the db and sample config params match
        CheckExistingConfig(database_config_file, config_sample)

        # if aDNA set new defaults
        anc_def = {"mapdamage": True, "min_prob": 0.50, "mismatch_probability": 0.20} if args.aDNA else {}

        # add all command line options to the merged config
        config = {**self.config_merged, **database_config, **sample_config, **anc_def, **vars(args)}

        # check if a db is built from an older version of haystac
        if isinstance(config["refseq_rep"], bool):
            raise ValidationError(
                "You are trying to use a database that was built with an older version of HAYSTAC. "
                "Please rebuild your database with the latest version."
            )

        target_list = list()

        if args.mode == "filter":
            if config["read_mode"] == PE:
                target_list.append(f"fastq/PE/{config['sample_prefix']}_mapq_pair.readlen")
            else:
                target_list.append(f"fastq/{config['read_mode']}/{config['sample_prefix']}_mapq.readlen")

        elif args.mode == "align":
            target_list.append(f"alignments/{config['sample_prefix']}_{config['read_mode']}_alignments.done")

        elif args.mode == "likelihoods":
            target_list.append(
                f"probabilities/{config['sample_prefix']}/{config['sample_prefix']}_likelihood_ts_tv_matrix.csv"
            )

        elif args.mode == "probabilities":
            target_list.append(
                f"probabilities/{config['sample_prefix']}/{config['sample_prefix']}_posterior_probabilities.tsv"
            )

        elif args.mode == "abundances":
            target_list.append(
                f"probabilities/{config['sample_prefix']}/{config['sample_prefix']}_posterior_abundance.tsv"
            )

        elif args.mode == "reads":
            target_list.append(f"dirichlet_reads/{config['sample_prefix']}_dirichlet_reads.done")

        # if mapdamage is true add it to the target list
        if config["mapdamage"]:
            target_list.append(f"mapdamage/{config['sample_prefix']}_mapdamage.done")

        config_analysis = os.path.join(args.analysis_output_dir, config["sample_prefix"] + "_config.yaml")

        if os.path.exists(config_analysis):
            CheckExistingConfig(config_analysis, config)

        with open(config_analysis, "w") as fout:
            yaml.safe_dump(config, fout, default_flow_style=False)

        target_list = [os.path.join(args.analysis_output_dir, target) for target in target_list]

        return self._run_snakemake("analyse", args, config, target_list)

    def _common_arguments(self, parser):
        """
        Add the common arguments shared by the `database`, `sample` and `analyse` commands
        """
        common = parser.add_argument_group("Common arguments")

        # add the help option manually so we can control where it is shown in the menu
        common.add_argument("-h", "--help", action="help", help="Show this help message and exit")
        common.add_argument("-v", "--version", action="help", help="Print the version number and exit")

        common.add_argument(
            "--cores",
            help="Maximum number of CPU cores to use",
            metavar="<int>",
            type=IntRangeType(1, MAX_CPU),
            default=MAX_CPU if self.config_default["cores"] == "all" else self.config_default["cores"],
        )

        common.add_argument(
            "--mem",
            help="Maximum memory (MB) to use",
            type=IntRangeType(1024, MAX_MEM_MB),
            default=MAX_MEM_MB if self.config_default["mem"] == "all" else self.config_default["mem"],
            metavar="<int>",
        )

        common.add_argument(
            "--unlock",
            help="Unlock the output directory following a crash or hard restart",
            action="store_true",
        )

        common.add_argument(
            "--debug",
            help="Enable debugging mode",
            action="store_true",
        )

        common.add_argument(
            "--snakemake",
            help="Pass additional flags to the `snakemake` scheduler.",
            metavar="'<json>'",
            type=JsonType(),
        )

    def _run_snakemake(self, module, args, config, target_list):
        """
        Helper function for running the snakemake workflow
        """
        print(f"HAYSTAC v {__version__}\n")
        print(f"Date: {datetime.datetime.now()}\n")

        config["module"] = module
        config["workflow_dir"] = os.path.join(CODE_DIR, "workflow")

        print("Config parameters:\n")
        params = config if args.debug else vars(args)

        for key, value in params.items():
            if value or args.debug:
                print(f" {key}: {value}")
        print("\n")

        if args.debug:
            print("Target files:\n")
            for target in target_list:
                print(" " + target)
            print("\n")

        os.makedirs(SNAKE_DIR, exist_ok=True)

        # save the run-time config file
        with open(CONFIG_RUNTIME, "w") as fout:
            yaml.safe_dump(config, fout, default_flow_style=False)

        # get any extra snakemake params
        smk_params = config.pop("snakemake") or {}

        # get any rule targets for batching and convert them into a compatible object
        exec_batch = Batch(config["batch"][0], config["batch"][1], config["batch"][2]) if config.get("batch") else None
        if config.get("batch"):
            config.pop("batch")

        success = snakemake.snakemake(
            snakefile=os.path.join(CODE_DIR, "workflow/workflow.smk"),
            config=config,
            targets=target_list,
            batch=exec_batch,
            cores=int(args.cores),
            resources={"entrez_api": self.max_entrez_requests, "mem_mb": int(args.mem)},
            force_incomplete=True,
            scheduler="greedy",
            # handle the rule-specific conda environments
            use_conda=config["use_conda"],
            conda_frontend="mamba",
            conda_prefix=os.path.join(config["cache"], "conda") if config["use_conda"] else None,
            # set all the debugging flags
            printreason=args.debug,
            printshellcmds=args.debug,
            show_failed_logs=args.debug,
            verbose=args.debug,
            keep_incomplete=args.debug,
            restart_times=0 if args.debug else RESTART_TIMES,
            keepgoing=(not args.debug),
            unlock=args.unlock,
            # pass on any CLI arguments from the --snakemake flag
            **smk_params,
        )

        # tidy up all the snakemake metadata
        if success and (len(os.listdir(os.path.join(SNAKE_DIR, "locks"))) == 0):
            shutil.rmtree(SNAKE_DIR)

        # translate "success" into shell exit code of 0
        return 0 if success else 1


if __name__ == "__main__":
    Haystac()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import pandas as pd

from haystac.workflow.scripts.entrez_utils import entrez_assembly_ftp
from haystac.workflow.scripts.utilities import REGEX_BLACKLIST


def entrez_refseq_create_files(
    config,
    input_file,
    nuccore_genomes_out,
    genbank_genomes_out,
    assemblies_out,
    nuccore_plasmids_out,
    genbank_plasmids_out,
):
    prok_refseq_rep = pd.read_csv(input_file, sep="\t")

    prok_refseq_rep_rmdup = prok_refseq_rep[~prok_refseq_rep["#Species/genus"].duplicated(keep="first")]

    assemblies = prok_refseq_rep_rmdup.loc[prok_refseq_rep_rmdup["WGS"].notna(), ["#Species/genus", "WGS"]]

    assemblies["#Species/genus"] = assemblies["#Species/genus"].replace(REGEX_BLACKLIST, "_", regex=True)

    nuccore = prok_refseq_rep_rmdup.loc[
        prok_refseq_rep_rmdup["Chromosome RefSeq"].notna(),
        ["#Species/genus", "Chromosome RefSeq"],
    ]

    nuccore["#Species/genus"] = nuccore["#Species/genus"].replace(REGEX_BLACKLIST, "_", regex=True)

    genbank = prok_refseq_rep_rmdup.loc[
        prok_refseq_rep_rmdup["Chromosome GenBank"].notna(),
        ["#Species/genus", "Chromosome GenBank"],
    ]

    genbank["#Species/genus"] = genbank["#Species/genus"].replace(REGEX_BLACKLIST, "_", regex=True)

    genbank_filtered = genbank[
        (~genbank["#Species/genus"].isin(assemblies["#Species/genus"]))
        & (~genbank["#Species/genus"].isin(nuccore["#Species/genus"]))
    ].copy()
    genbank_filtered.loc[:, "Chromosome GenBank"] = genbank_filtered["Chromosome GenBank"].str.split(",")
    genbank_exploded = genbank_filtered.explode("Chromosome GenBank")
    genbank_exploded = genbank_exploded[~genbank_exploded["#Species/genus"].duplicated(keep="first")]

    nuccore_filtered = nuccore[
        (~nuccore["#Species/genus"].isin(assemblies["#Species/genus"]))
        & (~nuccore["#Species/genus"].isin(genbank_filtered["#Species/genus"]))
    ].copy()
    nuccore_filtered.loc[:, "Chromosome RefSeq"] = nuccore_filtered["Chromosome RefSeq"].str.split(",")
    nuccore_exploded = nuccore_filtered.explode("Chromosome RefSeq")
    nuccore_exploded = nuccore_exploded[~nuccore_exploded["#Species/genus"].duplicated(keep="first")]

    assemblies_filtered = assemblies[
        (~assemblies["#Species/genus"].isin(nuccore_filtered["#Species/genus"]))
        & (~assemblies["#Species/genus"].isin(genbank_filtered["#Species/genus"]))
    ]
    assemblies_filtered.loc[:, "WGS"] = assemblies_filtered["WGS"].str.split(",")
    assemblies_exploded = assemblies_filtered.explode("WGS")
    assemblies_exploded = assemblies_exploded[~assemblies_exploded["#Species/genus"].duplicated(keep="first")]

    nuccore_plasmids = prok_refseq_rep_rmdup[
        prok_refseq_rep_rmdup["Plasmid RefSeq"].notna() & prok_refseq_rep_rmdup["WGS"].isna()
    ].loc[:, ["#Species/genus", "Plasmid RefSeq"]]

    genbank_plasmids = prok_refseq_rep_rmdup[
        prok_refseq_rep_rmdup["Plasmid GenBank"].notna() & prok_refseq_rep_rmdup["WGS"].isna()
    ].loc[:, ["#Species/genus", "Plasmid GenBank"]]
    genbank_plasmids_filtered = genbank_plasmids[
        ~genbank_plasmids["#Species/genus"].isin(nuccore_plasmids["#Species/genus"])
    ].copy()

    nuccore_plasmids.loc[:, "Plasmid RefSeq"] = nuccore_plasmids["Plasmid RefSeq"].str.split(",")
    nuccore_plasmids_exploded = nuccore_plasmids.explode("Plasmid RefSeq")

    nuccore_plasmids_exploded["#Species/genus"] = nuccore_plasmids_exploded["#Species/genus"].replace(
        REGEX_BLACKLIST, "_", regex=True
    )

    genbank_plasmids_filtered.loc[:, "Plasmid GenBank"] = genbank_plasmids_filtered["Plasmid GenBank"].str.split(",")
    genbank_plasmids_filtered_exploded = genbank_plasmids_filtered.explode("Plasmid GenBank")

    genbank_plasmids_filtered_exploded["#Species/genus"] = genbank_plasmids_filtered_exploded["#Species/genus"].replace(
        REGEX_BLACKLIST, "_", regex=True
    )

    header = ["species", "AccessionVersion"]

    if config["sequences"] or config["accessions"]:
        user_inputs = []
        if os.path.isfile(config["sequences"]):
            custom_fasta_paths = pd.read_csv(
                config["sequences"],
                sep="\t",
                header=None,
                names=["species", "accession", "path"],
            )
            user_inputs.append(custom_fasta_paths)
        if os.path.isfile(config["accessions"]):
            custom_accessions = pd.read_csv(
                config["accessions"],
                sep="\t",
                header=None,
                names=["species", "accession"],
            )
            user_inputs.append(custom_accessions)

        for user_df in user_inputs:
            genbank_exploded = genbank_exploded[(~genbank_exploded["#Species/genus"].isin(user_df["species"]))]
            nuccore_exploded = nuccore_exploded[(~nuccore_exploded["#Species/genus"].isin(user_df["species"]))]
            assemblies_exploded = assemblies_exploded[(~assemblies_exploded["#Species/genus"].isin(user_df["species"]))]
            genbank_plasmids_filtered_exploded = genbank_plasmids_filtered_exploded[
                (~genbank_plasmids_filtered_exploded["#Species/genus"].isin(user_df["species"]))
            ]
            nuccore_plasmids_exploded = nuccore_plasmids_exploded[
                (~nuccore_plasmids_exploded["#Species/genus"].isin(user_df["species"]))
            ]

    genbank_plasmids_to_drop = []
    for key, seq in genbank_plasmids_filtered_exploded.iterrows():
        if key in genbank_plasmids_to_drop:
            continue
        accession = seq["Plasmid GenBank"]
        url = entrez_assembly_ftp(accession, config["force_accessions"])
        if url != "":
            genbank_plasmids_to_drop.append(key)
    genbank_plasmids_filtered_exploded.drop(genbank_plasmids_to_drop, inplace=True)

    nuccore_plasmids_to_drop = []
    for key, seq in nuccore_plasmids_exploded.iterrows():
        if key in nuccore_plasmids_to_drop:
            continue
        accession = seq["Plasmid RefSeq"]
        url = entrez_assembly_ftp(accession, config["force_accessions"])
        if url != "":
            nuccore_plasmids_to_drop.append(key)
    nuccore_plasmids_exploded.drop(nuccore_plasmids_to_drop, inplace=True)

    genbank_exploded.to_csv(genbank_genomes_out, sep="\t", header=header, index=False)
    nuccore_exploded.to_csv(nuccore_genomes_out, sep="\t", header=header, index=False)
    assemblies_exploded.to_csv(assemblies_out, sep="\t", header=header, index=False)
    genbank_plasmids_filtered_exploded.to_csv(genbank_plasmids_out, sep="\t", header=header, index=False)
    nuccore_plasmids_exploded.to_csv(nuccore_plasmids_out, sep="\t", header=header, index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    entrez_refseq_create_files(
        config=snakemake.config,
        input_file=snakemake.input[0],
        nuccore_genomes_out=snakemake.output[0],
        genbank_genomes_out=snakemake.output[1],
        assemblies_out=snakemake.output[2],
        nuccore_plasmids_out=snakemake.output[3],
        genbank_plasmids_out=snakemake.output[4],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import numpy as np
import pandas as pd


def calculate_taxa_probabilities(ts_tv_matrix_file, params_file, sample_name, total_fastq_reads, output_file):
    """
    Function that calculates the species identification posterior probabilities.
    This function acts as a wrapper around the calculate_probabilities function.
    Function calculate_probabilities actually calculates the probabilities.
    Function calculate_taxa_probabilities acts as a wrapper by providing the right dataset each time.
    """
    assert os.stat(ts_tv_matrix_file).st_size, f"The ts_tv count file is empty {ts_tv_matrix_file}"
    assert os.stat(params_file).st_size, f"The probability model parameters file is empty {params_file}"

    print("all taxa", "\t", sample_name, file=sys.stderr)

    calculate_probabilities(
        ts_tv_matrix_file,
        params_file,
        sample_name,
        total_fastq_reads,
        output_file,
        submatrix="all taxa",
    )


def calculate_probabilities(
    ts_tv_matrix_file,
    params_file,
    sample_name,
    total_fastq_reads,
    output_file,
    submatrix,
):
    total_fastq_reads = float(open(total_fastq_reads, "r").read())

    read_count = len(pd.read_csv(ts_tv_matrix_file, sep=",", usecols=["Read_ID"])["Read_ID"].unique())

    if not read_count:
        print("File is empty, moving on.")

    ts_tv_matrix = pd.read_csv(ts_tv_matrix_file, sep=",", usecols=["Taxon", "Ts", "Tv"])

    model_params = pd.read_json(params_file, orient="index").squeeze()

    mismatch_df = ts_tv_matrix.groupby("Taxon").agg(
        {
            "Ts": lambda num: (sum(num) + ((read_count - num.count()) * (model_params["ts_missing_val"] + 1))),
            "Tv": lambda num: (sum(num) + ((read_count - num.count()) * (model_params["tv_missing_val"] + 1))),
        }
    )

    mismatch_df = mismatch_df.astype({"Ts": float, "Tv": float})
    mismatch_df.reset_index(inplace=True)
    mismatch_df["Posterior"] = np.nan
    mismatch_df["Read_Count"] = read_count
    mismatch_df["Total_Reads"] = total_fastq_reads
    mismatch_df["Submatrix"] = submatrix
    mismatch_df["Sample"] = sample_name

    for index, row in mismatch_df.iterrows():
        mismatch_t_vec = mismatch_df["Ts"].subtract(row["Ts"]).rpow(model_params["delta_t"])
        mismatch_v_vec = mismatch_df["Tv"].subtract(row["Tv"]).rpow(model_params["delta_v"])
        products = mismatch_t_vec * mismatch_v_vec
        denominator_summation = products.sum()
        posterior_row = float(1) / denominator_summation
        mismatch_df.iloc[index, 3] = posterior_row

    mismatch_df.sort_values(by="Posterior", ascending=False, inplace=True)

    print(mismatch_df, file=sys.stderr)

    with open(output_file, "a") as output_handle:
        mismatch_df.to_csv(path_or_buf=output_handle, sep="\t", index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    calculate_taxa_probabilities(
        ts_tv_matrix_file=snakemake.input[0],
        params_file=snakemake.input[1],
        sample_name=snakemake.wildcards.sample,
        total_fastq_reads=snakemake.input[2],
        output_file=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import random
import sys


def random_db_paths(input_list, output, seed):
    """Function that randomises the genomes that are going to be split into chunks for the bt2 index."""

    random.Random(seed).shuffle(input_list)

    with open(output, "w") as fout:
        for path in input_list:
            fout.write(f"{path}\n")


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    random_db_paths(
        input_list=snakemake.input,
        output=snakemake.output[0],
        seed=snakemake.params[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import pandas as pd

from haystac.workflow.scripts.utilities import REGEX_BLACKLIST


def entrez_pick_sequences(config, nuccore_file, taxa_file, output_file):
    accessions = pd.read_csv(nuccore_file, sep="\t")
    taxa = pd.read_csv(taxa_file, sep="\t")
    rank = config["rank"]

    sequences = pd.merge(accessions, taxa, on="TaxId", how="outer")

    sequences = sequences[~sequences[rank].isnull()]

    selected_sequences = sequences.loc[
        sequences.groupby(rank)["Length"].idxmax(),
        ["species", "AccessionVersion"],
    ]

    if config["refseq_rep"]:
        refseq_genomes = pd.read_csv(config["db_output"] + "/entrez/refseq-genomes.tsv", sep="\t")
        genbank_genomes = pd.read_csv(config["db_output"] + "/entrez/genbank-genomes.tsv", sep="\t")
        assemblies = pd.read_csv(config["db_output"] + "/entrez/assemblies.tsv", sep="\t")
        refseq_plasmids = pd.read_csv(config["db_output"] + "/entrez/refseq-plasmids.tsv", sep="\t")
        genbank_plasmids = pd.read_csv(config["db_output"] + "/entrez/genbank-plasmids.tsv", sep="\t")

        # the entrez query might give a different accession for a certain species than the refseq rep one and
        # I don't want that. If the species exists in the refseq I want to keep the refseq records
        selected_sequences = selected_sequences[
            ~selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True).isin(refseq_genomes.species)
        ]
        selected_sequences = selected_sequences[
            ~selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True).isin(genbank_genomes.species)
        ]
        selected_sequences = selected_sequences[
            ~selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True).isin(assemblies.species)
        ]
        selected_sequences = selected_sequences[
            ~selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True).isin(refseq_plasmids.species)
        ]
        selected_sequences = selected_sequences[
            ~selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True).isin(genbank_plasmids.species)
        ]

    selected_sequences["species"] = selected_sequences["species"].replace(REGEX_BLACKLIST, "_", regex=True)

    if config["sequences"]:
        custom_fasta_paths = pd.read_csv(
            config["sequences"],
            sep="\t",
            header=None,
            names=["species", "accession", "path"],
        )

        selected_sequences = selected_sequences[(~selected_sequences["species"].isin(custom_fasta_paths["species"]))]

    if config["accessions"]:
        custom_accessions = pd.read_csv(
            config["accessions"],
            sep="\t",
            header=None,
            names=["species", "accession"],
        )

        selected_sequences = selected_sequences[(~selected_sequences["species"].isin(custom_accessions["species"]))]

    selected_sequences.to_csv(output_file, sep="\t", header=True, index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    entrez_pick_sequences(
        config=snakemake.config,
        nuccore_file=snakemake.input[0],
        taxa_file=snakemake.input[1],
        output_file=snakemake.output[0],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os

import pandas as pd

from haystac.workflow.scripts.utilities import print_error


def calculate_bt2_idx_chunks(mem_resources, mem_rescale_factor, fasta_files, output_tsv, output_txt):
    """Calculate the number of chunks that the db sequences are going to be split into"""

    fasta_paths_random = []
    with open(str(fasta_files), "r") as fin:
        for line in fin:
            fasta_paths_random.append(line.strip())

    chunk_size = float(mem_resources) / float(mem_rescale_factor)
    total_size = 0.0
    chunks = 1

    chunk_df_list = []

    for fasta_file in fasta_paths_random:
        file_size = os.stat(fasta_file).st_size / (1024 ** 2)

        if file_size >= mem_resources or file_size >= chunk_size:
            print_error(
                f"Fasta file {fasta_file} is bigger than the RAM resources provided. "
                f"Unfortunately an index cannot be built."
            )

        if total_size + file_size >= chunk_size:
            total_size = file_size
            chunks += 1
        else:
            total_size += file_size

        chunk_df_list.append([chunks, fasta_file])

    chunk_df = pd.DataFrame(chunk_df_list, columns=["chunk", "path"])

    chunk_df.to_csv(output_tsv, sep="\t", index=False, header=False)

    idx_chunk_total = chunk_df["chunk"].max()

    with open(output_txt, "w") as outfile:
        print(idx_chunk_total, file=outfile)


if __name__ == "__main__":

    # noinspection PyUnresolvedReferences
    calculate_bt2_idx_chunks(
        mem_resources=snakemake.params[1],
        mem_rescale_factor=snakemake.params[2],
        fasta_files=snakemake.input,
        output_tsv=snakemake.output[0],
        output_txt=snakemake.output[1],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import shutil
import sys


# TODO get rid of this file and do the concat operations on the command line using `cat {list} | xargs cat > {output}`
def concat_files(list_of_files, output_file):
    """Concatenating files that are provided as input into a single file."""

    print("Concatenating files ...", file=sys.stderr)

    with open(output_file, "w") as fout:
        for file in list_of_files:
            print(file, file=sys.stderr)
            with open(file, "r") as fin:
                shutil.copyfileobj(fin, fout)

    print(f"{output_file} file created", file=sys.stderr)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    concat_files(list_of_files=snakemake.input, output_file=snakemake.output[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Evangelos A. Dimopoulos, Evan K. Irving-Pease"
__copyright__ = "Copyright 2020, University of Oxford"
__email__ = "antonisdim41@gmail.com"
__license__ = "MIT"

import os
import sys

import numpy as np
import pandas as pd

from haystac.workflow.scripts.utilities import REGEX_BLACKLIST


def entrez_refseq_eukaryotes_create_files(
    config,
    input_file,
    euk_genomes_out,
):

    """Function to parse the refseq genomes report for eukaryotes."""

    # read the file

    refseq_eukaryotes = pd.read_csv(input_file, sep="\t")

    # drop duplicate species/strains and get pairs of taxa and a assembly accession codes

    refseq_eukaryotes_rmdup = refseq_eukaryotes[~refseq_eukaryotes["#Organism/Name"].duplicated(keep="first")]

    eukaryotes_unique = refseq_eukaryotes_rmdup[["#Organism/Name", "Assembly Accession"]].copy()

    # drop rows that have no accessions

    means_empty_record = ["", "-", np.nan]

    eukaryotes = eukaryotes_unique.loc[~eukaryotes_unique["Assembly Accession"].isin(means_empty_record)]

    # rename columns

    eukaryotes.rename(columns={"#Organism/Name": "species", "Assembly Accession": "AccessionVersion"}, inplace=True)

    # regex for species name

    eukaryotes["species"] = eukaryotes["species"].replace(REGEX_BLACKLIST, "_", regex=True)

    # check for duplicates from user input

    if config["sequences"] or config["accessions"]:
        user_inputs = []
        if os.path.isfile(config["sequences"]):
            custom_fasta_paths = pd.read_csv(
                config["sequences"],
                sep="\t",
                header=None,
                names=["species", "accession", "path"],
            )
            user_inputs.append(custom_fasta_paths)
        if os.path.isfile(config["accessions"]):
            custom_accessions = pd.read_csv(
                config["accessions"],
                sep="\t",
                header=None,
                names=["species", "accession"],
            )
            user_inputs.append(custom_accessions)

        for user_df in user_inputs:
            eukaryotes = eukaryotes[(~eukaryotes["species"].isin(user_df["species"]))]

    # print the output to csv

    header = ["species", "AccessionVersion"]
    eukaryotes.to_csv(euk_genomes_out, sep="\t", header=header, index=False)


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    sys.stderr = open(snakemake.log[0], "w")

    # noinspection PyUnresolvedReferences
    entrez_refseq_eukaryotes_create_files(
        config=snakemake.config,
        input_file=snakemake.input[0],
        euk_genomes_out=snakemake.output[0],
    )
